"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

# -*- coding: utf-8 -*-
import sys
import functools
import datetime
import threading
import traceback
from libs import base
from flask import json
#from flask import current_app
from flask_login import session
#from flask_login import current_user
from models import cluster
from models.user import User
from constants import *
from helpers.synchronizing import synchronous
from werkzeug.exceptions import Unauthorized
from utilities.config_manager import config_manager


class ClusterManager(object):
    """It manages the all current cluster objects"""

    def __init__(self):

        self._clusters = {}
        self.clusters_lock = threading.RLock()

    @synchronous('clusters_lock')
    def create_cluster(self, node_address, user=None, password=None):

        cluster_info = self.get_cluster_data_by_node(node_address)
        cluster_obj = self.get_cluster_obj_by_node(node_address)

        if cluster_info is None:
            if type(node_address) is str:
                ip, port = node_address.split(":")
                return cluster.Cluster(ip, port, user=user, password=password)
            else:
                return EXISTING_CLUSTER

        """Create new cluster for different user same cluster"""
        if user not in cluster_obj.get_user():
            if type(node_address) is str:
                ip, port = node_address.split(":")
                return cluster.Cluster(ip, port, user=user, password=password)
        if user is not None:
            ip, port = node_address.split(":")
            return cluster.Cluster(ip, port, user=user, password=password)
        return EXISTING_CLUSTER

    def add_cluster(self, cluster):

        self._clusters[cluster.get_cluster_id()] = cluster

    def remove_cluster(self, cluster_id):

        del(self._clusters[cluster_id])

    def get_cluster_by_id(self, cluster_id):

        return self._clusters.get(cluster_id, None)

    def get_cluster_data_by_cluster_object(self, cluster_obj):

        return dict(cluster_id=cluster_obj.get_cluster_id(),
            nodes=cluster_obj.get_node_list().values(),
            seed_address=cluster_obj.get_seed(),
            security_enabled=cluster_obj.is_secured(),
            update_interval=config_manager.get_update_interval())

    def get_cluster_data_by_node(self, node_address):

        for cluster_id, cluster_obj in list(self._clusters.items()):
            if node_address in cluster_obj.get_nodes():
                return self.get_cluster_data_by_cluster_object(cluster_obj)
        return None

    def get_cluster_data_by_cluster_id(self, cluster_key):

        for cluster_id, cluster_obj in list(self._clusters.items()):
            if cluster_id == cluster_key:
                return self.get_cluster_data_by_cluster_object(cluster_obj)
        return None

    def get_cluster_obj_by_node(self, node_address):

        for cluster_id, cluster_obj in list(self._clusters.items()):
            if node_address in cluster_obj.get_nodes():
                return cluster_obj
        return None

    def get_cluster_info(self, host_address, user=None, password=None):

        try:
            ip, port = host_address.split(":")
        except:
            return dict(error="Unable to connect - %s" % host_address)

        is_security_enabled = base.is_security_enabled(ip, port)

        if is_security_enabled:
            node_address = base.get_self_address(ip, port, user=user, password=password)
        else:
            node_address = base.get_self_address(ip, port)

        if node_address == UNABLE_TO_CONNECT:
            return dict(error="Unable to connect - %s" % host_address)
        elif user and password and node_address in [UNRECOGNISED_USER, BAD_PASSWORD, BAD_CREDENTIAL]:
            raise Unauthorized("Unauthenticated user!")
        elif node_address in [UNRECOGNISED_USER, BAD_PASSWORD, BAD_CREDENTIAL]:
            return dict(security_enabled=is_security_enabled, cluster_id=None)
        cluster_info = None
        if not is_security_enabled:
            cluster_info = self.get_cluster_data_by_node(node_address)
        if cluster_info is not None:

            cluster_obj = self.get_cluster_obj_by_node(node_address)
            """cluster_obj.resume()
            # TODO: Add current user to cluster object's users' dict
            # DONE
            if is_security_enabled:
                user_obj = User(
                    user, password, host=host_address
                    )
                #cluster_obj = self.get_cluster_obj_by_node(node_address)
                cluster_obj.add_user(user_obj)
            return cluster_info"""

            """Same Cluster Different User check"""

            if user in cluster_obj.get_user():

                cluster_obj.resume()
                # TODO: Add current user to cluster object's users' dict
                # DONE
                if is_security_enabled:
                    user_obj = User(
                        user, password, host=host_address
                        )
                    #cluster_obj = self.get_cluster_obj_by_node(node_address)
                    cluster_obj.add_user(user_obj)

                return cluster_info

        if not is_security_enabled:
            cluster = self.create_cluster(node_address)
        if is_security_enabled and user is not None:
            cluster = self.create_cluster(node_address, user=user, password=password)


        if cluster == UNABLE_TO_CONNECT or cluster == ERROR_OCCURRED:
            return dict(error="Unable to connect - %s" % host_address)
        elif cluster == EXISTING_CLUSTER:
            cluster_info = self.get_cluster_data_by_node(node_address)
            if cluster_info is not None:
                # TODO: Add current user to cluster object's users' dict
                # DONE
                if is_security_enabled:
                    user_obj = User(
                        user, password, host=node_address
                        )
                    cluster_obj = self.get_cluster_obj_by_node(node_address)
                    cluster_obj.add_user(user_obj)
                return cluster_info
        else:
            self.add_cluster(cluster)
            return self.get_cluster_data_by_cluster_object(cluster)

    def get_clusters(self):

        return self._clusters

    def remove_unused_clusters(self):

        current_time = datetime.datetime.now()

        for cluster_id, cluster_obj in list(self._clusters.items()):
            elapsed_time = current_time - cluster_obj.get_last_access_time()
            if elapsed_time.seconds > CLUSTER_TIMEOUT * 60:
                cluster_obj.pause()
                #self.remove_cluster(cluster_id)

    def check_cluster_validity(self):

        def decorator(fn):

            @functools.wraps(fn)
            def f(*args, **kwargs):

                #if str(kwargs['cluster']) != session.get('cluster_id', ''):
                    #return json.dumps(dict(error="Invalid cluster id : %s" % kwargs["cluster"]))

                cluster_id = session.get('cluster_id', None) or kwargs.get('cluster', None)

                cluster = self.get_cluster_by_id(cluster_id)
                if cluster is None:
                    return json.dumps(dict(error="Invalid cluster id : %s" % kwargs["cluster"]))

                cluster.update_last_access_time()
                kwargs["cluster"] = cluster

                #kwargs['user'] = user
                #kwargs['password'] = password

                try:
                    return fn(*args, **kwargs)
                except Unauthorized as ua:
                    return ua
                except Exception:
                    sys.stderr.write("args:%s\nkwargs:%s" % (args, kwargs))
                    sys.stderr.write((traceback.format_exc()))
                    raise Exception("Internal server error!")

            return f
        return decorator
